﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;
using HeuristicLab.Core;
using HeuristicLab.Common;
using System.Reflection;

namespace HeuristicLab.Parameters {
  [StorableClass]
  public class ParameterSecondLevelNameBinding : IParameterBinding {
    [Storable]
    private IValueParameter source;
    public IValueParameter Source {
      get { return source; }
    }
    [Storable]
    private ILookupParameter target;
    public ILookupParameter Target {
      get { return target; }
    }
    [Storable]
    private string sourcePropertyName;
    public string SourcePropertyName {
      get { return sourcePropertyName; }
    }
    [Storable]
    private ParameterNameBinding nameBinding;

    [StorableConstructor]
    protected ParameterSecondLevelNameBinding(bool deserializing) { }
    protected ParameterSecondLevelNameBinding(ParameterSecondLevelNameBinding original, Cloner cloner) {
      this.source = cloner.Clone(source);
      this.target = cloner.Clone(target);
      this.sourcePropertyName = original.sourcePropertyName;
      RegisterEventHandlers(); // nameBinding does not need to be cloned, is recreated here
    }
    public ParameterSecondLevelNameBinding(ILookupParameter target, IValueParameter source, string sourcePropertyName) {
      if (source == null) throw new ArgumentNullException("source");
      if (target == null) throw new ArgumentNullException("target");
      this.source = source;
      this.target = target;
      this.sourcePropertyName = sourcePropertyName;
      this.nameBinding = null;
    }

    public object Clone() {
      return Clone(new Cloner());
    }

    public virtual IDeepCloneable Clone(Cloner cloner) {
      return new ParameterSecondLevelNameBinding(this, cloner);
    }

    public void Bind() {
      RegisterEventHandlers();
    }

    public void Unbind() {
      DeregisterEventHandlers();
    }

    [StorableHook(HookType.AfterDeserialization)]
    private void AfterDeserialization() {
      RegisterEventHandlers();
    }

    private void RegisterEventHandlers() {
      source.ValueChanged += new EventHandler(source_ValueChanged);
      IParameter sourceParameter = GetNestedParameter(source, sourcePropertyName);
      if (sourceParameter != null) {
        nameBinding = new ParameterNameBinding(target, sourceParameter);
        nameBinding.Bind();
      }
    }

    private void DeregisterEventHandlers() {
      source.ValueChanged -= new EventHandler(source_ValueChanged);

      if (nameBinding != null) nameBinding.Unbind();
      nameBinding = null;
    }

    private void source_ValueChanged(object sender, EventArgs e) {
      IParameter sourceParameter = GetNestedParameter(source, sourcePropertyName);
      if (sourceParameter != null) {
        if (nameBinding != null) nameBinding.Unbind();
        nameBinding = new ParameterNameBinding(target, sourceParameter);
        nameBinding.Bind();
      }
    }

    private IParameter GetNestedParameter(IValueParameter param, string name) {
      if (param.Value == null) return null;
      PropertyInfo property = param.Value.GetType().GetProperty(name, BindingFlags.Instance | BindingFlags.NonPublic);
      if (property == null) return null;
      if (property.GetGetMethod() == null) throw new InvalidOperationException("Property " + name + " of type " + param.Value.GetType().FullName + " does not have a Get method.");
      IParameter nestedParam = (property.GetGetMethod().Invoke(param.Value, null) as IParameter);
      return nestedParam;
    }
  }
}
